--[[
-- Roomba
- (Thanks to BalkiTheWise for the name)
 ]]

Roomba = Roomba or {}
Roomba.DELAY = 100
Roomba.format = "                     %3d"
Roomba.currentRun = {}
Roomba.checkingBank = false
Roomba.restackInProgress = false

local defaults = {}
defaults.RoombaAtGBank = true
defaults.RoombaAtBank = true
defaults.RoombaInBag = true
defaults.RoombaPosition = KEYBIND_STRIP_ALIGN_LEFT

-- Return InstanceID of an item at slot X in a bag
function Roomba.GetInstanceId(bag, slotId)

	for i, v in pairs(bag) do
		if v.slotIndex == slotId then
			return v.itemInstanceId
		end
	end
	
	return nil
	
end

-- Flag for other addons. Returns true while Roomba restacks
function Roomba.WorkInProgress()
	return Roomba.restackInProgress
end

-- Confirm slot position of a slot or return nil
function Roomba.FindSlot(bag, slotId)
	
	for i, v in pairs(bag) do
		-- Found our slot
		if v.slotIndex == slotId then return true end
	end
	
	return nil
	
end

-- Count entrys in a table - TODO: check if #table works
function Roomba.TableEntryCount(t)
	local i = next(t)
	local c = 0
	while i do
		c = c + 1
		i = next(t, i)
	end
	return c
end

-- Return true if they're 5 slots left in the inventory
function Roomba.CheckWeHaveEnoughRoom()
	return CheckInventorySpaceAndWarn(5)
end

-- Bank is ready! Find those duplicates! Runs when EVENT_GUILD_BANK_ITEMS_READY + 1s
function Roomba.RoombaReady()

	--d("Roomba.RoombaReady()")
	
	-- Protect
	if ZO_GuildBankBackpack:IsHidden() then return end
	
	-- Are we in the process of checking the bank? Need to protect due to Keybinding
	if not Roomba.checkingBank then return end
	
	local bagToScan = SHARED_INVENTORY:GenerateFullSlotData(nil, BAG_GUILDBANK)
	
	-- If Guild does not get a bank , should not happend
	if DoesGuildHavePrivilege(Roomba.currentBank, GUILD_PRIVILEGE_BANK_DEPOSIT) then
		-- If no permission, don't do
		if DoesPlayerHaveGuildPermission(Roomba.currentBank, GUILD_PERMISSION_BANK_DEPOSIT) and DoesPlayerHaveGuildPermission(Roomba.currentBank, GUILD_PERMISSION_BANK_WITHDRAW) then
			
			Roomba.scanInStackableBag(bagToScan)
			
			-- If they're is no buttons, add them
			if not KEYBIND_STRIP:HasKeybindButtonGroup(Roomba.descriptor) then
				if not KEYBIND_STRIP[Roomba.keyBindCheck] then
					KEYBIND_STRIP:AddKeybindButtonGroup(Roomba.descriptor)
				elseif not KEYBIND_STRIP[Roomba.keyBindCheck][2] then
					Roomba.keyBindIndex = 2
					KEYBIND_STRIP:AddKeybindButtonGroup(Roomba.descriptor)
				end
			-- Update descriptors. Descriptors update will call Roomba.HaveStuffToStack and show Restack button if needed
			elseif KEYBIND_STRIP:HasKeybindButtonGroup(Roomba.descriptor) then
				KEYBIND_STRIP:UpdateKeybindButtonGroup(Roomba.descriptor)
			end
			
			Roomba.currentRun = {}
			
		else
			if KEYBIND_STRIP:HasKeybindButtonGroup(Roomba.descriptor) then
				KEYBIND_STRIP:RemoveKeybindButtonGroup(Roomba.descriptor) 
			end
		end
	end
	
end

function Roomba.ResetAll()

	Roomba.cInstanceId = nil
	Roomba.cSlotIdx = nil
	Roomba.cSlot = nil
	Roomba.inBagCollection = {}
	Roomba.currentReturnIndex = nil
	Roomba.restackInProgress = false
	Roomba.keyBindIndex = 1
	Roomba.waitingRetries = 1
	
end

function Roomba.SelectGuildBank(_, guildBankId)
	
	-- Reset flag for bank switch
	Roomba.checkingBank = false
	Roomba.currentBank = guildBankId
	Roomba.ResetAll()
	
end

function Roomba.ReturnItemsToBank(_, errorCode)
	
	--d("Roomba.ReturnItemsToBank()")
	
	-- Protect for fast Escape while we restack
	if ZO_GuildBankBackpack:IsHidden() then
		Roomba.stopGBRestackAndRestartScan()
		return false
	end
	
	-- Now that we've stacked it all lets move it back
	
	-- guildbank restack in progress
	if Roomba.gBankRestack then
		-- If we made a restack just before
		if Roomba.lastRestackResult then
		
			--d("Roomba.ReturnItemsToBank()->lastRestackResult")
			
			if errorCode == GUILD_BANK_NO_SPACE_LEFT then
			
				-- Stop. Guild Bank is full, User need to clean it manually
				
				--d("Roomba.ReturnItemsToBank()->errorCode:" .. GUILD_BANK_NO_SPACE_LEFT)
				
				Roomba.stopGBRestackAndRestartScan()
				return false
			
			-- Can occur if an addon has destroyed our item while we were restacking
			elseif errorCode == GUILD_BANK_ITEM_NOT_FOUND then
				
				--d("Roomba.ReturnItemsToBank()->errorCode:" .. GUILD_BANK_ITEM_NOT_FOUND)
				-- Need to move to next stack
				
				-- Let's try next try of the same item. Is there another stack to move?
				if next(Roomba.lastRestackResult[Roomba.itemIndex], Roomba.slotIndex) then
					
					-- yes, try to move it
					Roomba.slotIndex = Roomba.slotIndex + 1
					
					zo_callLater(Roomba.ReturnItemsToBank, Roomba.DELAY)
					
				-- No more stack. Maybe another item?
				elseif next(Roomba.lastRestackResult, Roomba.itemIndex) then
					-- Yes, move its first stack
					Roomba.itemIndex = Roomba.itemIndex + 1
					Roomba.slotIndex = 1
					
					zo_callLater(Roomba.ReturnItemsToBank, Roomba.DELAY)
					
				else
					-- Nothing to move
					return
				end
			-- Can occur if an addon has destroyed our item while we were restacking
			elseif errorCode == GUILD_BANK_TRANSFER_PENDING then
				
				--d("Roomba.ReturnItemsToBank()->errorCode:" .. GUILD_BANK_TRANSFER_PENDING)
				-- Retry to move the stack
				
				Roomba.waitingRetries = Roomba.waitingRetries + 1
				if Roomba.waitingRetries < 10 then
					RoombaWindowDescription:SetText("Guild bank busy, trying to return restacked " .. Roomba.cSlot.name .. " to the Guild Bank again")
					zo_callLater(Roomba.ReturnItemsToBank, Roomba.DELAY * 5)
				else
					Roomba.stopGBRestackAndRestartScan()
				end
			
			-- No errors, is our item here ?
			elseif Roomba.FindSlot(SHARED_INVENTORY:GenerateFullSlotData(nil, BAG_BACKPACK), Roomba.lastRestackResult[Roomba.itemIndex][Roomba.slotIndex].slot) then
				RoombaWindowDescription:SetText("Returning restacked " .. Roomba.cSlot.name .. " to the Guild Bank")
				TransferToGuildBank(BAG_BACKPACK, Roomba.lastRestackResult[Roomba.itemIndex][Roomba.slotIndex].slot)
				-- It will trigger .BankItemsReceived because of EVENT_GUILD_BANK_ITEM_ADDED
				-- It will trigger .ReturnItemsToBank if an error has occured
			-- Another error
			elseif errorCode then
				--d("Roomba.ReturnItemsToBank->errorCode:" .. errorCode)
				-- Not handled yet, stop
				return
			else
				
				-- It's a 3rd party addon push while we were restacking, or item has been destroyed
				
				--d("Roomba.ReturnItemsToBank()->itemDestroyed")
				Roomba.restackGuildbank()
				
			end
		end
	end
	
end

-- Triggers when EVENT_GUILD_BANK_ITEM_ADDED
function Roomba.BankItemsReceived(_, gslot)
	
	-- guildbank restack in progress
	if Roomba.gBankRestack then
		-- If we made a restack just before
		if Roomba.lastRestackResult then

			-- Roomba is restacking the guild bank
			
			-- Is the item added our last move ?
			-- Get its instanceID
			local id = Roomba.GetInstanceId(SHARED_INVENTORY:GenerateFullSlotData(nil, BAG_GUILDBANK), gslot)
			
			-- Protection
			if id ~= Roomba.lastRestackResult[Roomba.itemIndex][Roomba.slotIndex].itemInstanceId then return end
			
			--If we're here that's because it's our Roomba item which was sent in GuildBank
			
			-- Let's move the next stack
			-- Let's try next try of the same item. Is there another stack to move?
			if next(Roomba.lastRestackResult[Roomba.itemIndex], Roomba.slotIndex) then
				
				-- yes, try to move it
				Roomba.slotIndex = Roomba.slotIndex + 1
				
				zo_callLater(Roomba.ReturnItemsToBank, Roomba.DELAY)
				
			-- No more stack. Maybe another item? - Should not happen , because there is only 1 item in .lastRestackResult, the others items are in Roomba.duplicates
			elseif next(Roomba.lastRestackResult, Roomba.itemIndex) then
			
				-- Yes, move its first stack
				Roomba.itemIndex = Roomba.itemIndex + 1
				Roomba.slotIndex = 1
				
				zo_callLater(Roomba.ReturnItemsToBank, Roomba.DELAY)
				
			else
			
				-- Nothing else to move for this item, let's do the rest
				--d("Roomba.BankItemsReceived->finished")
				Roomba.restackGuildbank()
				
			end
			
		end
	end
	
end

-- Event called that item has arrived
function Roomba.receiveItemInBagpack(_, bagId, slotId)

	--d("Roomba.receiveItemInBagpack()")
	
	-- Protection
	if bagId ~= BAG_BACKPACK then return end
	
	-- Set in Roomba.restackGuildbank
	if not Roomba.cItemDuplicateList then return end
	
	-- useful ?
	if IsItemJunk(BAG_BACKPACK, slotId) then
		SetItemIsJunk(BAG_BACKPACK, slotId, false)
		zo_callLater(Roomba.receiveItemInBagpack, 1000)	
	end
	
	local bagBackPack = SHARED_INVENTORY:GenerateFullSlotData(nil, BAG_BACKPACK)
	
	-- Is slot really used?
	if not Roomba.GetInstanceId(bagBackPack, slotId) then return end
	
	-- Is slot == our item transferred, avoid manual transfers interfer while we restack
	if Roomba.GetInstanceId(bagBackPack, slotId) ~= Roomba.cSlot.itemInstanceId then return end
	
	-- Set in wich bagpackslot our stack is, and overwrite (guildbank) slot position by the bagpack one
	Roomba.cSlot.bagpackSlot = slotId
	Roomba.cSlot.slot = slotId
	
	-- Build an array for merging in backpack
	table.insert(Roomba.inBagCollection, Roomba.cSlot)
	
	-- If we have another slot to move from GuildBank
	if next(Roomba.cItemDuplicateList, Roomba.cSlotIdx) then
		
		Roomba.cSlotIdx, Roomba.cSlot = next(Roomba.cItemDuplicateList, Roomba.cSlotIdx)
		
		local bagGuildBank = SHARED_INVENTORY:GenerateFullSlotData(nil, BAG_GUILDBANK)
		
		-- Is slot really used?
		if not Roomba.GetInstanceId(bagGuildBank, Roomba.cSlot.slot) then
			Roomba.stopGBRestackAndRestartScan()
			return false
		end
		
		-- Is slot == our item transferred, avoid manual transfers interfer while we restack
		if Roomba.GetInstanceId(bagGuildBank, Roomba.cSlot.slot) ~= Roomba.cSlot.itemInstanceId then
			Roomba.stopGBRestackAndRestartScan()
			return false
		end
		
		--d("Roomba.receiveItemInBagpack()->Transfering stack #" .. Roomba.cSlotIdx)
		-- The TransferFromGuildBank will execute .receiveItemInBagpack because of registration
		TransferFromGuildBank(Roomba.cSlot.slot)
		
	else 
		
		-- No more slots to move, let's stack them
		-- Disable Roomba.receiveItemInBagpack, we've finished to transfer all stacks of same item
		EVENT_MANAGER:UnregisterForEvent("RoombaInventoryAdded", EVENT_INVENTORY_SINGLE_SLOT_UPDATE)
		
		-- Stack items in Bagpack
		zo_callLater(function()
						
						RoombaWindowDescription:SetText("Stacking " .. Roomba.cSlot.name .. " in inventory") -- TODO: Handle user leave bank while restack. (need to do this on all zo_callLater)
						-- restack items Add an array up to our array to cheat
						Roomba.restackStackableBag(BAG_BACKPACK, {Roomba.inBagCollection}, true)
						
						--Init
						Roomba.itemIndex = 1
						Roomba.slotIndex = 1
						
						-- These events will loop the move back to the guild bank
						-- This event is only here to "retry" if a return has fail
						EVENT_MANAGER:RegisterForEvent("RoombaGuildBankError", EVENT_GUILD_BANK_TRANSFER_ERROR, Roomba.ReturnItemsToBank)
						-- This event will triger the next transfer to bagpack?
						EVENT_MANAGER:RegisterForEvent("RoombaGuildBankSuccess", EVENT_GUILD_BANK_ITEM_ADDED, Roomba.BankItemsReceived)
						-- Return the stack(s)
						
						zo_callLater(Roomba.ReturnItemsToBank, Roomba.DELAY * 5)
						
					end, Roomba.DELAY)
		
	end
	
end

-- Restack the bank
function Roomba.restackGuildbank()

	--d("Roomba.restackGuildbank()")
	
	-- Protect
	if ZO_GuildBankBackpack:IsHidden() then return end
	
	-- 5 slots Needed to Work
	if not Roomba.CheckWeHaveEnoughRoom() then return end
	
	-- Resetted after each restack / each bank swap
	if Roomba.cInstanceId == nil then
		Roomba.checkingBank = false
		Roomba.RoombaReady()
		Roomba.checkingBank = true
	end
	
	-- ?
	ZO_MenuBar_SelectLastVisibleButton(ZO_PlayerInventoryTabs)
	
	-- Pull the next job off the stack
	Roomba.cInstanceId, Roomba.cItemDuplicateList = next(Roomba.duplicates, Roomba.cInstanceId)
	
	-- Protect against Keybind
	if Roomba.cInstanceId then Roomba.currentRun[Roomba.cInstanceId] = true end
	
	--d("Roomba.restackGuildbank()->ok")
	
	-- They are some duplicates .cItemDuplicateList is an array of multiple stacks of same item
	if Roomba.cItemDuplicateList then
		
		-- Flag for other addons
		Roomba.restackInProgress = true
		
		-- Show progress Bar, etc.
		RoombaWindow:SetHidden(false)
		local index = Roomba.TableEntryCount(Roomba.currentRun)
		local total = Roomba.TableEntryCount(Roomba.duplicates)
		ZO_StatusBar_SmoothTransition(RoombaWindowSpeedRow.bar, index , total, FORCE_VALUE)
		RoombaWindowSpeedRow.value:SetText(string.format(Roomba.format, (index / total) * 100) .. "%")
		
		--d("Roomba.restackGuildbank()->Working on item #" .. Roomba.cInstanceId)
		
		-- Init
		Roomba.cSlotIdx = 1
		
		-- 1st stack
		Roomba.cSlot = Roomba.cItemDuplicateList[Roomba.cSlotIdx]
		
		-- cSlot is the slot to get from GuildBank
		RoombaWindowIcon:SetTexture(Roomba.cSlot.texture)
		
		-- Init what's actually in our bagpack
		Roomba.inBagCollection = {}
		
		-- If it suddenly doesn't exist, try the next in the list (can be caused by addons which autodestroy or other transfering utilities).
		if not Roomba.FindSlot(SHARED_INVENTORY:GenerateFullSlotData(nil, BAG_GUILDBANK), Roomba.cSlot.slot) then return zo_callLater(Roomba.restackGuildbank, Roomba.DELAY) end
		
		-- Will trigger the function when TransferFromGuildBank will be executed
		EVENT_MANAGER:RegisterForEvent("RoombaInventoryAdded", EVENT_INVENTORY_SINGLE_SLOT_UPDATE, Roomba.receiveItemInBagpack)
		
		RoombaWindowDescription:SetText("Retrieving " .. Roomba.cSlot.name .. " from Guild Bank")
		
		--d("Roomba.restackGuildbank()->Transfering stack #" .. Roomba.cSlotIdx)
		-- Take a stack from bank, it will trigger Roomba.receiveItemInBagpack
		TransferFromGuildBank(Roomba.cSlot.slot)
		
	else
	
		Roomba.stopGBRestackAndRestartScan()
		return false
		
	end
	
	return true
	
end

-- Stop a Guild bank restack and do another scan
function Roomba.stopGBRestackAndRestartScan()

	-- Finish or protection
	
	--d("Roomba.stopGBRestackAndRestartScan()")
	
	EVENT_MANAGER:UnregisterForEvent("RoombaGuildBankError", EVENT_GUILD_BANK_TRANSFER_ERROR)
	EVENT_MANAGER:UnregisterForEvent("RoombaGuildBankSuccess", EVENT_GUILD_BANK_ITEM_ADDED)
	EVENT_MANAGER:UnregisterForEvent("RoombaInventoryAdded", EVENT_INVENTORY_SINGLE_SLOT_UPDATE)
	
	-- Kick off the next transaction
	Roomba.ResetAll()
		
	-- Now rescan and show/hide roomba button
	RoombaWindowDescription:SetText("Complete")
	RoombaWindow:SetHidden(true)
	
	-- Perform another scan if an addon played with GuildBank while we restacking
	Roomba.RoombaReady()

end

function Roomba.onBankReady()
	
	if Roomba.opts.RoombaAtBank then
		Roomba.keyBindIndex = 1
		
		-- Bank is evented to be ready, but wait a short while before processing. Bank in fact got a 100/300ms lag when opening it
		zo_callLater(function() Roomba.onBagReady(BAG_BANK) end, 400)
		
		zo_callLater(function()
			if not KEYBIND_STRIP:HasKeybindButtonGroup(Roomba.descriptor) then
				--d("Roomba.displayBagButtons()->duplicates->add")
				if not KEYBIND_STRIP[Roomba.keyBindCheck] then
					KEYBIND_STRIP:AddKeybindButtonGroup(Roomba.descriptor)
				elseif not KEYBIND_STRIP[Roomba.keyBindCheck][2] then
					Roomba.keyBindIndex = 2
					KEYBIND_STRIP:AddKeybindButtonGroup(Roomba.descriptor)
				end
			end
		end, 500)
	end
	
end

function Roomba.onBagReady(bagId)

	-- Which bag to scan?
	local bagToScan = nil
	if bagId == BAG_BACKPACK then
		bagToScan = SHARED_INVENTORY:GenerateFullSlotData(nil, BAG_BACKPACK)
	elseif bagId == BAG_BANK then
		bagToScan = SHARED_INVENTORY:GenerateFullSlotData(nil, BAG_BANK)
	end
	
	Roomba.scanInStackableBag(bagToScan)
	
	-- Update descriptors
	if KEYBIND_STRIP:HasKeybindButtonGroup(Roomba.descriptor) then
		--d("Roomba.onBagReady()->duplicates->update")
		KEYBIND_STRIP:UpdateKeybindButtonGroup(Roomba.descriptor)
	end
	
end

function Roomba.bagNeedRestack()

	-- Protect
	if SCENE_MANAGER and SCENE_MANAGER:GetCurrentScene() then
		if SCENE_MANAGER:GetCurrentScene():GetName() ~= "guildBank" then
			if SCENE_MANAGER:GetCurrentScene():GetName() ~= "bank" then
				if SCENE_MANAGER:GetCurrentScene():GetName() ~= "inventory" then
					return false
				end
			end
		end
	end
	
	if Roomba.duplicates then
	
		local indexItemsDuplicated = next(Roomba.duplicates)
		
		-- There are some duplicates
		if indexItemsDuplicated then
			--d("Roomba.bagNeedRestack()->true")
			return true
		else
			--d("Roomba.bagNeedRestack()->false")
			return false
		end
		
	else
		return false
	end

end

function Roomba.bagRestack(bagId)

	--d("Roomba.bagRestack()")
	
	local indexItemsDuplicated = next(Roomba.duplicates)
	
	-- There are some duplicates
	if indexItemsDuplicated then
		
		Roomba.restackInProgress = true
		
		-- Stack them
		Roomba.restackStackableBag(bagId, Roomba.duplicates, false)
		
		Roomba.restackInProgress = false
		
	end

end

function Roomba.onBankClosed()

	if Roomba.opts.RoombaAtBank then
		--d("Roomba.bankClosed()")
		
		-- Hide buttons
		if KEYBIND_STRIP:HasKeybindButtonGroup(Roomba.descriptor) then
			KEYBIND_STRIP:RemoveKeybindButtonGroup(Roomba.descriptor) 
		end
	end

end

-- Scan in a stackable bag
function Roomba.scanInStackableBag(bagToScan)

	--d("Roomba.scanInStackableBag()")

	Roomba.lookUp = {}
	Roomba.duplicates = {}
	local duplitemp = {}
	
	-- We only need to store slots with items
	for index, slot in pairs(bagToScan) do
		
		-- Stack at max?
		Roomba.stack, Roomba.maxStack = GetSlotStackSize(slot.bagId, slot.slotIndex)
		
		-- Stack is not at max
		if Roomba.stack ~= Roomba.maxStack then
			
			-- REAL itemId
			Roomba.id = slot.itemInstanceId
			
			-- We already find this item before
			if Roomba.lookUp[Roomba.id] then
				-- Already marked as duplicate?
				if not duplitemp[Roomba.id] then
					--d("Found duplicate (" .. slot.name .. ")")
					-- Duplicate
					duplitemp[Roomba.id] = Roomba.lookUp[Roomba.id]
				end
			else
				-- New item found
				Roomba.lookUp[Roomba.id] = {}
			end
			
			-- Now group all items by id
			table.insert(Roomba.lookUp[Roomba.id], {slot = slot.slotIndex, stack = Roomba.stack, texture = slot.iconFile, name = slot.name, itemInstanceId = slot.itemInstanceId})
			
		end
		
	end
	
	local i = 1
	for _, data in pairs(duplitemp) do
		Roomba.duplicates[i] = data
		i = i + 1
	end

end

function Roomba.restackStackableBag(bagId, duplicateList, gBankRestack)

	--d("Roomba.restackStackableBag(" .. bagId .. ")")
	
	local restartAfter = false
	local itemIndex = 0
	local indexItemsDuplicated, dataItems = next(duplicateList)
	local result = {}
	Roomba.lastRestackResult = {}
	
	-- Loop for item X/Y/Z
	while indexItemsDuplicated do
		
		local index = 1
		local itemInfo = dataItems[index]
		local baseSlot = nil
		local lastMoveWasSingleStack = false
		local lastMoveWasMultiStack = false
		result[indexItemsDuplicated] = {}
		
		--d(".restackStackableBag-> " .. indexItemsDuplicated)
		
		-- Loop for item X and slots 1/2/3/y
		while itemInfo do
		
			--d(".restackStackableBag->checking index " .. index)
			
			-- 1st loop : we go in
			if baseSlot == nil then
				-- Our actual stack / Our max size
				baseSlot = itemInfo
				baseSlot.actualStack, baseSlot.maxStack = GetSlotStackSize(bagId, itemInfo.slot)
			else
				
				local qty
				-- If stacking, will we get 1 or 2 stacks ?
				itemInfo.maxStack = baseSlot.maxStack
				
				if (baseSlot.actualStack + itemInfo.stack) <= baseSlot.maxStack then
					
					-- Only 1 stack, we can merge stacks
					qty = itemInfo.stack
					
					-- Merging
					--d(".restackStackableBag->Merging slot " .. itemInfo.slot .. " into " .. baseSlot.slot)
					if IsProtectedFunction("RequestMoveItem") then
						CallSecureProtected("RequestMoveItem", bagId, itemInfo.slot, bagId, baseSlot.slot, qty)
					else
						RequestMoveItem(bagId, itemInfo.slot, bagId, baseSlot.slot, qty)
					end
					
					-- Update values baseSlot can still be used in next loop
					baseSlot.actualStack = baseSlot.actualStack + itemInfo.stack
					baseSlot.stack = baseSlot.actualStack
					
					if lastMoveWasMultiStack == true then
						result[indexItemsDuplicated][#result[indexItemsDuplicated]] = baseSlot
						lastMoveWasSingleStack = true
						lastMoveWasMultiStack = false
					elseif lastMoveWasSingleStack == true then
						result[indexItemsDuplicated][#result[indexItemsDuplicated]] = baseSlot
						lastMoveWasMultiStack = false
					else
						table.insert(result[indexItemsDuplicated], baseSlot)
						lastMoveWasSingleStack = true
						lastMoveWasMultiStack = false
					end
					
				else
					
					-- It won't fit, just move the qty to match maxStack, no need to rescan because slots don't move, only stacks.
					qty = baseSlot.maxStack - baseSlot.actualStack
					
					-- Merging
					--d("restackStackableBag->Merging slot " .. itemInfo.slot .. " into " .. baseSlot.slot)
					if IsProtectedFunction("RequestMoveItem") then
						CallSecureProtected("RequestMoveItem", bagId, itemInfo.slot, bagId, baseSlot.slot, qty)
					else
						RequestMoveItem(bagId, itemInfo.slot, bagId, baseSlot.slot, qty)
					end
					
					if lastMoveWasMultiStack == true then
						result[indexItemsDuplicated][#result[indexItemsDuplicated]].stack = baseSlot.maxStack
						result[indexItemsDuplicated][#result[indexItemsDuplicated]].actualStack = baseSlot.maxStack
						table.insert(result[indexItemsDuplicated], itemInfo)
						result[indexItemsDuplicated][#result[indexItemsDuplicated]].stack = itemInfo.stack - qty
						result[indexItemsDuplicated][#result[indexItemsDuplicated]].actualStack = itemInfo.stack - qty
						lastMoveWasSingleStack = false
					elseif lastMoveWasSingleStack == true then
						result[indexItemsDuplicated][#result[indexItemsDuplicated]].stack = baseSlot.maxStack
						result[indexItemsDuplicated][#result[indexItemsDuplicated]].actualStack = baseSlot.maxStack
						table.insert(result[indexItemsDuplicated], itemInfo)
						result[indexItemsDuplicated][#result[indexItemsDuplicated]].stack = itemInfo.stack - qty
						result[indexItemsDuplicated][#result[indexItemsDuplicated]].actualStack = itemInfo.stack - qty
						lastMoveWasSingleStack = false
						lastMoveWasMultiStack = true
					else
						table.insert(result[indexItemsDuplicated], baseSlot)
						result[indexItemsDuplicated][#result[indexItemsDuplicated]].stack = baseSlot.maxStack
						result[indexItemsDuplicated][#result[indexItemsDuplicated]].actualStack = baseSlot.maxStack
						table.insert(result[indexItemsDuplicated], itemInfo)
						result[indexItemsDuplicated][#result[indexItemsDuplicated]].stack = itemInfo.stack - qty
						result[indexItemsDuplicated][#result[indexItemsDuplicated]].actualStack = itemInfo.stack - qty
						lastMoveWasSingleStack = false
						lastMoveWasMultiStack = true
					end
					
					-- Baseslot is now at max, we cannot use it anymore
					baseSlot = itemInfo
					baseSlot.actualStack = itemInfo.stack - qty
					
				end
				
			end
			
			index = index + 1
			itemInfo = dataItems[index]
		
		end
		
		itemIndex = itemIndex + 1
		indexItemsDuplicated, dataItems = next(duplicateList, itemIndex)
		
	end
	
	--d("Roomba.restackStackableBag->finished")
	Roomba.lastRestackResult = result
	Roomba.gBankRestack = gBankRestack
	
	-- Rescan. Because Bank got a lag, we need to call Roomba.onBankReady() to avoid have our restacking job still unknown
	if bagId == BAG_BANK then
		Roomba.onBankReady()
	-- Only on restacking bag, not guildbank
	elseif bagId == BAG_BACKPACK and not gBankRestack then
		Roomba.onSceneStateChange(SCENE_HIDDEN, SCENE_SHOWING)
	end
	
end

function Roomba.onSceneStateChange(oldState, newState)
	
	if Roomba.opts.RoombaInBag then
		
		Roomba.keyBindIndex = 1
		
		if newState == SCENE_SHOWING then
			
			zo_callLater(function() Roomba.onBagReady(BAG_BACKPACK) end, 400)
			
			zo_callLater(function()
							if not KEYBIND_STRIP:HasKeybindButtonGroup(Roomba.descriptor) then
								--d("Roomba.displayBagButtons()->duplicates->add")
								if not KEYBIND_STRIP[Roomba.keyBindCheck] then
									KEYBIND_STRIP:AddKeybindButtonGroup(Roomba.descriptor)
								elseif not KEYBIND_STRIP[Roomba.keyBindCheck][2] then
									Roomba.keyBindIndex = 2
									KEYBIND_STRIP:AddKeybindButtonGroup(Roomba.descriptor)
								end
							end
						end, 500)
			
		end

		if newState == SCENE_HIDDEN then
			-- Hide buttons
			if KEYBIND_STRIP:HasKeybindButtonGroup(Roomba.descriptor) then
				KEYBIND_STRIP:RemoveKeybindButtonGroup(Roomba.descriptor) 
			end
		end
	end
	
end

-- RegisterCallback for inventory restacking
-- This declaration has to be set AFTER Roomba.onSceneStateChange declaration
Roomba.inventoryScene = SCENE_MANAGER:GetScene("inventory")
Roomba.inventoryScene:RegisterCallback("StateChange", Roomba.onSceneStateChange)

-- Triggers when EVENT_ADD_ON_LOADED
function Roomba.RoombaLoaded(_, addOnName)
	
	if(addOnName ~= "Roomba") then
		return
	end
	
	Roomba.descriptorName = "Run Roomba"
	ZO_CreateStringId("SI_BINDING_NAME_RUN_ROOMBA", Roomba.descriptorName)
	ZO_CreateStringId("ROOMBA_RESCAN_BAG", "Rescan Bag")
	ZO_CreateStringId("ROOMBA_RESCAN_BANK", "Rescan Bank")
	
	Roomba.InitialiseFrame()
	Roomba.InitialiseSettings()
	
	-- Set the function to run when guild bank is opened (before guild bank is ready)
	EVENT_MANAGER:RegisterForEvent("RoombaGuildBankOpen", EVENT_OPEN_GUILD_BANK, function()
																		if Roomba.opts.RoombaAtGBank then
																			Roomba.InitialiseEvents()
																		end
																	end)
	-- Set the function to run when guild bank is closed
	EVENT_MANAGER:RegisterForEvent("RoombaGuildBankClose", EVENT_CLOSE_GUILD_BANK, function()
																	-- If there is buttons, remove them
																	if Roomba.opts.RoombaAtGBank then
																		if KEYBIND_STRIP:HasKeybindButtonGroup(Roomba.descriptor) then
																			KEYBIND_STRIP:RemoveKeybindButtonGroup(Roomba.descriptor) 
																		end
																		RoombaWindow:SetHidden(true)
																		Roomba.ResetAll()
																		Roomba.UninitialiseEvents()
																		end
																	end)
	
end

function Roomba.InitialiseFrame()

	RoombaWindow:SetDrawLayer(DL_OVERLAY)
	RoombaWindowSpeedRowBarContainerValue:SetText(" 0%")
	
	ZO_StatusBar_SetGradientColor(RoombaWindowSpeedRow.bar, ZO_XP_BAR_GRADIENT_COLORS)
	ZO_StatusBar_SmoothTransition(RoombaWindowSpeedRow.bar, 0, 20, FORCE_VALUE)
	
	RoombaWindowSpeedRowIcon:SetHidden(true)
	
	RoombaWindowSpeedRowBarContainer:ClearAnchors()
	RoombaWindowSpeedRowBarContainer:SetAnchor(BOTTOM, RoombaWindowIcon, BOTTOM, 30, 70)
	
	RoombaWindowSpeedRowBarContainerValue:SetAnchor(BOTTOM, RoombaWindowIcon, BOTTOM, 140, 80)
	
	RoombaWindowSpeedRowBarContainerValue:SetFont("ZoFontHeader3")
	
end

function Roomba.isBagDisplayed()

	--d("Roomba.isBagDisplayed()")
	
	-- We can't "really" (?) update a keybind name dynamically while using "name = function() .. end" so we use the "visible" one. NEED TO RECHECK THIS
	
	if Roomba.bagNeedRestack() then
		Roomba.descriptorName = GetString(SI_BINDING_NAME_RUN_ROOMBA) 
	else
		
		if SCENE_MANAGER and SCENE_MANAGER:GetCurrentScene() then
			local scene = SCENE_MANAGER:GetCurrentScene():GetName()
			if scene == "bank" or scene == "guildBank" then
				Roomba.descriptorName = GetString(ROOMBA_RESCAN_BANK)
			elseif scene == "inventory" then
				Roomba.descriptorName = GetString(ROOMBA_RESCAN_BAG)
			end
		end
		
	end

	local scene = SCENE_MANAGER:GetCurrentScene():GetName()
	if scene == "bank" or scene == "inventory" or scene == "guildBank" then
		return true
	else
		return false
	end

end

function Roomba.InitialiseSettings()

	Roomba.currentBank = 1
	Roomba.keyBindIndex = 1
	
	-- Fetch the saved variables
	Roomba.opts = ZO_SavedVars:NewAccountWide('ROOMBA_OPTS', 1, nil, defaults)
	
	if Roomba.opts.RoombaPosition == KEYBIND_STRIP_ALIGN_LEFT then
		Roomba.keyBindCheck = "leftButtons"
	elseif Roomba.opts.RoombaPosition == KEYBIND_STRIP_ALIGN_RIGHT then
		Roomba.keyBindCheck = "rightButtons"
	end
	
	local panelData = {
		type = "panel",
		name = "Roomba",
		displayName = ZO_HIGHLIGHT_TEXT:Colorize("Roomba"),
		author = "Wobin, CrazyDutchGuy & Ayantir",
		version = 3,
		slashCommand = "/roomba",
		registerForRefresh = true,
		registerForDefaults = true,
	}
	
	local LAM = LibStub('LibAddonMenu-2.0')
	LAM:RegisterAddonPanel("RoombaOptions", panelData)
	
	local optionsTable = {}
	local index = 0
	
	index = index + 1
	optionsTable[index] = {
		type = "checkbox",
		name = Roomba.lang.RoombaAtGBank,
		tooltip = Roomba.lang.RoombaAtGBankTT,
		getFunc = function() return Roomba.opts.RoombaAtGBank end,
		setFunc = function(newValue) Roomba.opts.RoombaAtGBank = newValue end,
		width = "full",
		default = defaults.RoombaAtGBank,
	}
	
	index = index + 1
	optionsTable[index] = {
		type = "checkbox",
		name = Roomba.lang.RoombaAtBank,
		tooltip = Roomba.lang.RoombaAtBankTT,
		getFunc = function() return Roomba.opts.RoombaAtBank end,
		setFunc = function(newValue) Roomba.opts.RoombaAtBank = newValue end,
		width = "full",
		default = defaults.RoombaAtBank,
	}
	
	index = index + 1
	optionsTable[index] = {
		type = "checkbox",
		name = Roomba.lang.RoombaInBag,
		tooltip = Roomba.lang.RoombaInBagTT,
		getFunc = function() return Roomba.opts.RoombaInBag end,
		setFunc = function(newValue) Roomba.opts.RoombaInBag = newValue end,
		width = "full",
		default = defaults.RoombaInBag,
	}
	
	-- TODO : optimize
	index = index + 1
	optionsTable[index] = {
		type = "dropdown",
		name = Roomba.lang.RoombaPosition,
		tooltip = Roomba.lang.RoombaPositionTT,
		choices = {
			Roomba.lang.RoombaPositionChoice[KEYBIND_STRIP_ALIGN_LEFT],
			Roomba.lang.RoombaPositionChoice[KEYBIND_STRIP_ALIGN_RIGHT],
		},
		width = "full",
		default = defaults.RoombaPosition,
		warning = Roomba.lang.ReloadUI,
		getFunc = function() return Roomba.lang.RoombaPositionChoice[Roomba.opts.RoombaPosition] end,
		setFunc = function(choice)
			if choice == Roomba.lang.RoombaPositionChoice[KEYBIND_STRIP_ALIGN_LEFT] then
				Roomba.opts.RoombaPosition = KEYBIND_STRIP_ALIGN_LEFT
				Roomba.keyBindCheck = "leftButtons"
			elseif choice == Roomba.lang.RoombaPositionChoice[KEYBIND_STRIP_ALIGN_RIGHT] then
				Roomba.opts.RoombaPosition = KEYBIND_STRIP_ALIGN_RIGHT
				Roomba.keyBindCheck = "rightButtons"
			else
				-- When user click on LAM reinit button
				Roomba.opts.RoombaPosition = defaults.RoombaPosition
				Roomba.keyBindCheck = "leftButtons"
			end
			
			ReloadUI()
			
		end,
	}
	
	LAM:RegisterOptionControls("RoombaOptions", optionsTable)
	
	Roomba.descriptor = { 
		alignment = Roomba.opts.RoombaPosition,
		[Roomba.keyBindIndex] = {
			name = function() return Roomba.descriptorName end,
			keybind = "RUN_ROOMBA",
			control = self,
			callback = function(descriptor) Roomba.StartRoomba() end, 
			visible = function(descriptor) return Roomba.isBagDisplayed() end, 
			icon = [[Roomba\media\RoombaSearch.dds]],
		},
	}

end

-- Function mapped to the Keybinding
function Roomba.BeginStackingProcess()

	-- What function to use ? depends on the scene
	if SCENE_MANAGER:GetCurrentScene():GetName() == "inventory" and Roomba.opts.RoombaInBag then
		--RestackInventory
		Roomba.bagRestack(BAG_BACKPACK)
	elseif SCENE_MANAGER:GetCurrentScene():GetName() == "bank" and Roomba.opts.RoombaAtBank then
		-- Restack Bank
		Roomba.bagRestack(BAG_BANK)
	elseif SCENE_MANAGER:GetCurrentScene():GetName() == "guildBank" and Roomba.opts.RoombaAtGBank then
		-- Restack GuildBank
		Roomba.restackGuildbank()
	end

end

function Roomba.StartRoomba()

	if Roomba.bagNeedRestack() then
		Roomba.BeginStackingProcess()
	else
		Roomba.BeginScanningProcess()
	end

end

function Roomba.BeginScanningProcess()

	-- What function to use ? depends on the scene
	if SCENE_MANAGER:GetCurrentScene():GetName() == "inventory" and Roomba.opts.RoombaInBag then
		--RestackInventory
		Roomba.onBagReady(BAG_BACKPACK)
	elseif SCENE_MANAGER:GetCurrentScene():GetName() == "bank" and Roomba.opts.RoombaAtBank then
		-- Restack Bank
		Roomba.onBagReady(BAG_BANK)
	elseif SCENE_MANAGER:GetCurrentScene():GetName() == "guildBank" and Roomba.opts.RoombaAtGBank then
		-- Scan GuildBank
		Roomba.RoombaReady()
	end
	
end

function Roomba.onGuildBankReallyReady()

	-- Limit calls to Roomba.RoombaReady()
	if Roomba.checkingBank == false then
		Roomba.checkingBank = true
		Roomba.RoombaReady()
	end
	
end

function Roomba.onGuildBankReady()
	-- Guild bank is evented to be ready, but wait a short while before processing. (multiple readys for big banks ~3/4)
	zo_callLater(function() Roomba.onGuildBankReallyReady() end, 1500)
end

-- Runs when EVENT_OPEN_GUILD_BANK
function Roomba.InitialiseEvents()
	EVENT_MANAGER:RegisterForEvent("RoombaReady", EVENT_GUILD_BANK_ITEMS_READY, Roomba.onGuildBankReady)
	-- Clear the flag when swapping banks
	EVENT_MANAGER:RegisterForEvent("RoombaSelected", EVENT_GUILD_BANK_SELECTED, Roomba.SelectGuildBank)
end

-- Runs when EVENT_CLOSE_GUILD_BANK
function Roomba.UninitialiseEvents()
	EVENT_MANAGER:UnregisterForEvent("RoombaInventoryAdded", EVENT_INVENTORY_SINGLE_SLOT_UPDATE)
	EVENT_MANAGER:UnregisterForEvent("RoombaSelected", EVENT_GUILD_BANK_SELECTED)
	EVENT_MANAGER:UnregisterForEvent("RoombaReady", EVENT_GUILD_BANK_ITEMS_READY)
	EVENT_MANAGER:UnregisterForEvent("RoombaGuildBankError", EVENT_GUILD_BANK_TRANSFER_ERROR)
	EVENT_MANAGER:UnregisterForEvent("RoombaGuildBankSuccess", EVENT_GUILD_BANK_ITEM_ADDED)
	Roomba.checkingBank = false
end

EVENT_MANAGER:RegisterForEvent("Roomba", EVENT_ADD_ON_LOADED, Roomba.RoombaLoaded)

-- Set the function to run when bank is opened
EVENT_MANAGER:RegisterForEvent("RoombaBankOpen", EVENT_OPEN_BANK, Roomba.onBankReady)
																
-- Set the function to run when bank is closed
EVENT_MANAGER:RegisterForEvent("RoombaBankClosed", EVENT_CLOSE_BANK, Roomba.onBankClosed)